"""Session Manager"""

from abc import ABCMeta, abstractmethod
from datetime import datetime, timedelta, timezone
import logging
import secrets
from typing import (
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    TypeVar,
    TypedDict,
    cast
)

from bareasgi import HttpRequest
from bareutils import encode_set_cookie, header

LOGGER = logging.getLogger(__name__)


class Session(TypedDict):
    """The base session"""
    expiry: datetime


T = TypeVar('T', bound=Session)


class SessionManager(Generic[T], metaclass=ABCMeta):
    """The base class for session managers

    This is a cookie based session manager. When a client connects the cookie
    headers are checked to see if they contain a session cookie.

    If the session cookie doesn't exist a set-cookie header is added to the
    response headers containing a unique value to use as a session key. This
    cookie will then be sent by the client in all subsequent requests.

    If the session cookie is found the value of the cookie is used to access
    a session cache.
    """

    def __init__(self, session_duration: timedelta) -> None:
        """Initialize the session manager.

        Args:
            session_duration (timedelta): The time after which a session
                will expire.
        """
        self.session_duration = session_duration

        self._sessions: Dict[bytes, T] = {}
        # The name of the cookie is only relevant for the life of the session
        # manager.
        self._session_cookie_name = secrets.token_urlsafe().encode('ascii')

    def _get_session_key_from_cookie(
        self,
        request: HttpRequest
    ) -> Optional[bytes]:
        cookies = header.cookie(request.scope['headers'])
        session_cookie = cookies.get(
            self._session_cookie_name,
            [None]  # type: ignore
        )
        return next(iter(session_cookie))

    def _make_new_session_key(self) -> bytes:
        return secrets.token_hex(32).encode('ascii')

    @abstractmethod
    def create_session(self, expiry: datetime) -> T:
        """Create the session data.

        Args:
            expiry (datetime): The session expiry time.

        Returns:
            T: The session data.
        """

    def _make_session(self, now: datetime) -> Tuple[T, bytes]:
        """Make a new session.

        Args:
            now (datetime): The current time in UTC.

        Returns:
            Tuple[T, bytes]: The session data and the set-cookie value.
        """
        session_key = self._make_new_session_key()
        expiry = now + self.session_duration

        session = self.create_session(expiry)
        self._sessions[session_key] = session

        set_cookie = encode_set_cookie(
            self._session_cookie_name,
            session_key,
            expires=expiry
        )

        return session, set_cookie

    def _expire_sessions(self, now: datetime) -> None:
        """Delete all expired sessions.

        Args:
            now (datetime): The current time in UTC.
        """
        expired_session_keys = [
            key
            for key, session in self._sessions.items()
            if cast(Session, session)['expiry'] < now
        ]
        for key in expired_session_keys:
            del self._sessions[key]

    def get_session(
            self,
            request: HttpRequest
    ) -> Tuple[T, List[Tuple[bytes, bytes]]]:
        """Get an existing session, or create a new one.

        Args:
            request (HttpRequest): The HTTP request.

        Returns:
            Tuple[T, List[Tuple[bytes, bytes]]]: The session and an headers to
                send to maintain the session.
        """
        now = datetime.now(timezone.utc)

        self._expire_sessions(now)

        session_key = self._get_session_key_from_cookie(request)
        headers: List[Tuple[bytes, bytes]] = []

        session = (
            self._sessions.get(session_key)
            if session_key is not None
            else None
        )
        if session is None:
            session, set_cookie = self._make_session(now)
            headers.append((b'set-cookie', set_cookie))

        return session, headers



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bareasgi_sspi']

package_data = \
{'': ['*']}

install_requires = \
['bareasgi>=4.1.0,<5.0.0', 'pyspnego>=0.6.0,<0.7.0']

setup_kwargs = {
    'name': 'bareasgi-sspi',
    'version': '4.0.0rc1',
    'description': 'bareASGI middleware for SSPI',
    'long_description': '# bareASGI-sspi\n\n[ASGI](https://asgi.readthedocs.io/en/latest/index.html) middleware\nfor the [bareASGI](https://github.com/rob-blackbourn/bareASGI) framework\nproviding [SSPI](https://en.wikipedia.org/wiki/Security_Support_Provider_Interface) authentication\non Windows.\n\nThe implementation uses the [pyspnego](https://github.com/jborean93/pyspnego) package.\n\n## Installation\n\nInstall from the pie store.\n\n```\npip install bareasgi-sspi\n```\n\n## Usage\n\nThe following program uses the\n[Hypercorn](https://pgjones.gitlab.io/hypercorn/)\nASGI server, and the\n[bareASGI](https://github.com/rob-blackbourn/bareASGI)\nASGI framework.\n\n```python\nimport asyncio\nimport logging\nfrom typing import Optional\n\nfrom bareasgi import Application, HttpRequest, HttpResponse\nfrom bareutils import text_writer\nfrom hypercorn import Config\nfrom hypercorn.asyncio import serve\n\nfrom bareasgi_sspi import SPNEGOMiddleware, SSPIDetails\n\n# A callback to display the results of the SSPI middleware.\nasync def http_request_callback(request: HttpRequest) -> HttpResponse:\n    # Get the details from the request context request[\'sspi\']. Note if\n    # authentication failed this might be absent or empty.\n    sspi: Optional[SSPIDetails] = request.context.get(\'sspi\')\n    client_principal = (\n        sspi[\'client_principal\']\n        if sspi is not None\n        else \'unknown\'\n    )\n    return HttpResponse(\n        200,\n        [(b\'content-type\', b\'text/plain\')],\n        text_writer(f"Authenticated as \'{client_principal}\'")\n    )\n\n\nasync def main_async():\n    # Create the middleware. Change the protocol from Negotiate to NTLM,\n    # and allow unauthenticated requests to pass through.\n    sspi_middleware = SPNEGOMiddleware(\n        protocol=b\'NTLM\',\n        forbid_unauthenticated=False\n    )\n\n    # Make the ASGI application using the middleware.\n    app = Application(middlewares=[sspi_middleware])\n    app.http_router.add({\'GET\'}, \'/\', http_request_callback)\n\n    # Start the ASGI server.\n    config = Config()\n    config.bind = [\'localhost:9023\']\n    await serve(app, config)\n\nif __name__ == \'__main__\':\n    logging.basicConfig(level=logging.DEBUG)\n    asyncio.run(main_async())\n```\n\n### Arguments\n\nOptional arguments include:\n\n* `protocol` (`bytes`): Either `b"Negotiate"` or `b"NTLM"` (for systems not part of a domain).\n* `service` (`str`): The SPN service. Defaults to `"HTTP"`.\n* `hostname` (`str`, optional): The hostname. Defaults to he result of `socket.gethostname()`.\n* `session_duration` (`timedelta`, optional): The duration of a session. Defaults to 1 hour.\n* `forbid_unauthenticated` (`bool`): If true, and authentication fails, send 403 (Forbidden). Otherwise handle the request unauthenticated.\n* `context_key` (`str`, optional): The key used in the request context. Defaults to `sspi`.\n\n### Results\n\nIf the authentication is successful the SSPI details are added to the\n`context` dictionary of the HttpRequest object with the key `"sspi"`.\n\nThe following properties are set:\n\n* `"client_principal"` (`str`): The username of the client.\n* `"negotiated_protocol"` (`str`): The negotiated protocol.\n* `"protocol"` (`str`): The requested protocol.\n* `"spn"` (`str`): The SPN of the server.\n',
    'author': 'Rob Blackbourn',
    'author_email': 'rob.blackbourn@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rob-blackbourn/bareASGI-SSPI',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""bareASGI middleware for SSPI using spnego"""

import base64
from datetime import datetime, timedelta
import logging
import socket
from typing import (
    List,
    Literal,
    Optional,
    Tuple,
    TypedDict
)

from bareasgi import HttpRequest, HttpRequestCallback, HttpResponse
from bareutils import header, response_code
import spnego

from .session_manager import SessionManager, Session

LOGGER = logging.getLogger(__name__)


class SSPIDetails(TypedDict):
    """The details available from SPNEGO"""
    client_principal: str
    negotiated_protocol: str
    protocol: str
    spn: str


class SSPISession(Session):
    """The data to store for a session"""
    server_auth: spnego.ContextProxy
    status: Optional[Literal['requested', 'accepted', 'rejected']]
    sspi: Optional[SSPIDetails]


class SPNEGOSessionManager(SessionManager[SSPISession]):
    """The session manager implementation for SPNEGO"""

    def __init__(
            self,
            session_duration: timedelta,
            hostname: str,
            service: str,
            protocol: str
    ) -> None:
        super().__init__(session_duration)
        self.hostname = hostname
        self.service = service
        self.protocol = protocol

    def create_session(self, expiry: datetime) -> SSPISession:
        server_auth = spnego.server(
            hostname=self.hostname,
            service=self.service,
            protocol=self.protocol.lower()
        )

        session: SSPISession = {
            'server_auth': server_auth,
            'expiry': expiry,
            'status': None,
            'sspi': None
        }

        return session


class SPNEGOMiddleware:
    """ASGI middleware for authenticating with SSPI using SPNEGO

    Authentication data is stored in the scope `"extensions"` property under
    `"sspi"`.
    """

    def __init__(
            self,
            *,
            protocol: Literal[b'Negotiate', b'NTLM'] = b'Negotiate',
            service: str = 'HTTP',
            hostname: Optional[str] = None,
            session_duration: timedelta = timedelta(hours=1),
            forbid_unauthenticated: bool = True,
            context_key: str = 'sspi'
    ) -> None:
        """Initialise the SPNEGO middleware.

        Args:
            protocol (Literal[b&#39;Negotiate&#39;, b&#39;NTLM&#39;], optional):
                The protocol. Defaults to b'Negotiate'.
            service (str, optional): The service. Defaults to 'HTTP'.
            hostname (Optional[str], optional): The hostname. Defaults to None.
            session_duration (timedelta, optional): The duration of a session
                before re-authentication is performed. Defaults to
                `timedelta(hours=1)`.
            forbid_unauthenticated (bool, optional): If true, 403 (Forbidden) is
                sent if authentication fails. If false the request is handled,
                but no authentication details are added. Defaults to True.
            context_key (str, optional): The name of the key that will be used
                to store the data in the HttpRequest context.
                Defaults to 'sspi'.
        """
        self.protocol = protocol
        self.forbid_unauthenticated = forbid_unauthenticated
        self.context_key = context_key

        if hostname is None:
            hostname = socket.gethostname()

        self._session_manager = SPNEGOSessionManager(
            session_duration,
            hostname,
            service,
            self.protocol.decode('ascii')
        )

    def _forbidden(self) -> HttpResponse:
        return HttpResponse.from_text(
            'Forbidden',
            status=response_code.FORBIDDEN
        )

    def _unauthorized(self, headers: List[Tuple[bytes, bytes]]) -> HttpResponse:
        return HttpResponse.from_text(
            "Unauthorized",
            headers=headers,
            status=response_code.UNAUTHORIZED
        )

    async def _handle_accepted(
            self,
            session: SSPISession,
            request: HttpRequest,
            handler: HttpRequestCallback
    ) -> HttpResponse:
        # Store the result of the authentication in the request context.
        request.context[self.context_key] = session['sspi']
        return await handler(request)

    def _request_authentication(
            self,
            session: SSPISession,
            headers: List[Tuple[bytes, bytes]],
            request: HttpRequest
    ) -> HttpResponse:
        LOGGER.debug(
            "Requesting authentication for client %s using %s",
            request.scope['client'],
            self.protocol
        )

        # To request authentication a message with status 401 (Unauthorized)
        # is sent with the "www-authenticate" header set to the desired
        # protocol (e.g. Negotiate or NTLM). The client will then respond with
        # a message containing the "authorization" header. This header will
        # return the auth-scheme that the client has chosen, and a token.
        session['status'] = 'requested'
        return self._unauthorized([
            (b'www-authenticate', self.protocol),
            *headers
        ])

    async def _authenticate(
            self,
            authorization: Optional[bytes],
            session: SSPISession,
            request: HttpRequest,
            handler: HttpRequestCallback
    ) -> HttpResponse:
        # The authentication handshake involves an exchange of tokens.

        if not authorization:
            raise RuntimeError("Missing 'authorization' header")

        # Feed the client token into the authenticator. This is a two step
        # process.
        in_token = base64.b64decode(authorization[len(self.protocol)+1:])
        server_auth = session['server_auth']
        buf = server_auth.step(in_token)

        if server_auth.complete:
            # This is the second step. The handshake has succeeded and the
            # request can be passed on to the downstream handlers.
            LOGGER.debug(
                "Authentication succeeded for client %s using %s as user %s",
                request.scope['client'],
                self.protocol,
                server_auth.client_principal
            )
            session['status'] = 'accepted'
            session['sspi'] = {
                'client_principal': server_auth.client_principal,
                'protocol': server_auth.protocol,
                'negotiated_protocol': server_auth.negotiated_protocol,
                'spn': server_auth.spn
            }
            return await self._handle_accepted(session, request, handler)

        if buf:
            # This is the first step. The client has sent an acceptable token
            # and the server responds with its own as another 401 response.
            LOGGER.debug(
                "Sending challenge for client %s using %s",
                request.scope['client'],
                self.protocol
            )
            out_token = self.protocol + b" " + base64.b64encode(buf)
            return self._unauthorized([(b'www-authenticate', out_token)])

        raise RuntimeError("Handshake failed")

    async def _handle_requested(
            self,
            session: SSPISession,
            request: HttpRequest,
            handler: HttpRequestCallback
    ) -> HttpResponse:
        authorization = header.find(b'authorization', request.scope['headers'])
        try:
            return await self._authenticate(
                authorization,
                session,
                request,
                handler
            )
        except:  # pylint: disable=bare-except
            LOGGER.exception(
                "Failed to authenticate for client %s using %s",
                request.scope['client'],
                self.protocol
            )
            session['status'] = 'rejected'
            return await self._handle_rejected(session, request, handler)

    async def _handle_rejected(
            self,
            session: SSPISession,
            request: HttpRequest,
            handler: HttpRequestCallback
    ) -> HttpResponse:
        if self.forbid_unauthenticated:
            return self._forbidden()
        else:
            return await self._handle_accepted(session, request, handler)

    async def __call__(
            self,
            request: HttpRequest,
            handler: HttpRequestCallback
    ) -> HttpResponse:
        try:
            session, headers = self._session_manager.get_session(request)

            if session['status'] is None:
                return self._request_authentication(session, headers, request)
            elif session['status'] == 'requested':
                return await self._handle_requested(session, request, handler)
            elif session['status'] == 'accepted':
                return await self._handle_accepted(session, request, handler)
            elif session['status'] == 'rejected':
                return await self._handle_rejected(session, request, handler)
            else:
                raise RuntimeError("Unhandled session status")

        except:  # pylint: disable=bare-except
            LOGGER.exception("Failed to authenticate")
            return HttpResponse.from_text(
                "Internal Server Error",
                status=response_code.INTERNAL_SERVER_ERROR,
            )
